from tempfile import TemporaryDirectory
from distutils.dir_util import copy_tree
import shutil

import logging
import os
# from os.path import join
# from pathlib import Path

from mkdocs.utils import warning_filter

log = logging.getLogger(__name__)
log.addFilter(warning_filter)


class TfMerger:
    def __init__(self, config):
        self.config = config
        self.root_docs_dir = config['docs_dir']
        self.docs_dirs = list()
        self.files_source_dir = dict()

    def append(self, docs_path, abs_path):
        self.docs_dirs.append([docs_path, abs_path])

    def merge(self):
        self.temp_docs_dir = TemporaryDirectory('', 'docs_')
        # log.critical("temp_docs_dir = %s" % self.temp_docs_dir.name)
        # log.critical("root_docs_dir = %s" % self.root_docs_dir)

        # log.critical("ls root_docs_dir = %s" % list(Path(self.root_docs_dir).glob('**/*')))

        copy_tree(self.root_docs_dir, self.temp_docs_dir.name)

        # log.critical("ls temp_docs_dir = %s" % list(Path(self.temp_docs_dir.name).glob('**/*')))

        for docs_path, abs_path in self.docs_dirs:
            destination = os.path.join(
                self.temp_docs_dir.name, docs_path)
            destinationfolder = os.path.dirname(destination)
            if not os.path.exists(destinationfolder):
                os.makedirs(destinationfolder)
            shutil.copy(abs_path, destination)

        return str(self.temp_docs_dir.name)

    def getFilesSourceFolder(self):
        return self.files_source_dir

    def cleanup(self):
        return self.temp_docs_dir.cleanup()



from mkdocs.plugins import BasePlugin
from .parser import TfParser
from .merger import TfMerger
from collections import OrderedDict
# from pathlib import Path

import logging

from mkdocs.utils import warning_filter

log = logging.getLogger(__name__)
log.addFilter(warning_filter)

known_nav_special_syntax_plugins = ["monorepo"]


class TerraformMonorepoPlugin(BasePlugin):
    def __init__(self):
        self.parser = None
        self.merger = None
        self.originalDocsDir = None
        self.resolvedPaths = []
        self.files_source_dir = {}

    # need to add config for file readme and ignore regex list
    # would be good to add a check for terrafrom-docs html comment and remove it
    # before render event so we can use the page title in the nav

    def on_config(self, config):
        # If no 'nav' defined, we don't need to run.
        if not config.get('nav'):
            return config

        # terraform-monorepo has to be before monorepo in the plugins list
        # taken from https://github.com/timvink/mkdocs-enumerate-headings-plugin/blob/ef8b0931890b12455b53ccd52f41fcb3104eb89c/mkdocs_enumerate_headings_plugin/plugin.py#L46  # noqa: E501
        plugins = [*OrderedDict(config["plugins"])]

        for p in known_nav_special_syntax_plugins:
            if p in plugins:
                if plugins.index("terraform-monorepo") < plugins.index(p):
                    log.critical(
                        "[mkdocs-terraform-monorepo] terraform-monorepo should be defined after the %s plugin in your mkdocs.yml file"   # noqa: E501
                        % p
                    )
                    raise SystemExit(1)

        # Handle !tf_modules_root statements
        self.parser = TfParser(config)
        resolvedNav = self.parser.resolve()
        resolvedPaths = self.parser.getResolvedPaths()
        config['nav'] = resolvedNav

        # Generate a new "docs" directory
        self.merger = TfMerger(config)
        for docsPath, absPath in resolvedPaths:
            self.merger.append(docsPath, absPath)
        new_docs_dir = self.merger.merge()

        # Update the docs_dir with our temporary one!
        self.originalDocsDir = config['docs_dir']
        config['docs_dir'] = new_docs_dir

        # Store resolved paths for later.
        self.resolvedPaths = resolvedPaths

        # Store source directory of copied files for later
        self.files_source_dir = self.merger.getFilesSourceFolder()
        return config

    def on_pre_page(self, page, config, files):
        # Update page source attribute to point to source file
        # Only in case any files were moved.
        if len(self.files_source_dir) > 0:
            if page.file.abs_src_path in self.files_source_dir:
                page.file.abs_src_path = self.files_source_dir[page.file.abs_src_path]
        return page

    def on_serve(self, server, config, **kwargs):
        buildfunc = list(server.watcher._tasks.values())[0]['func']

        # still watch the original docs/ directory
        if self.originalDocsDir is not None:
            server.watch(self.originalDocsDir, buildfunc)

        # watch all the sub docs/ folders
        for watch_file, _ in self.resolvedPaths:
            server.watch(watch_file, buildfunc)

    def post_build(self, config):
        self.merger.cleanup()



import setuptools


setuptools.setup(
    name='mkdocs-terraform-monorepo-plugin',
    version='0.1.0',
    description='Plugin for adding doc tree for terraform monorepos.',
    long_description_content_type='text/plain',
    long_description="""
        This introduces support for the !tf_modules_root syntax in mkdocs.yml, allowing you to import trees of README.md files for terraform modules directories that also contains a \*.tf file.
        It enables large or complex repositories to have their own tree of modules folders, whilst generating only a single Mkdocs site.
        This is built and maintained by the engineering community at Wunderman Thompson Commerce.
        It was, however, inspired by and adapted from the Spotify backstage monorepo plugin.
    """,  # noqa: E501
    keywords='mkdocs terraform monorepo',
    url='https://github.com/wtc-cloud-eng/mkdocs-terraform-monorepo-plugin',
    author='Richard Wittrick',
    author_email='richard.wittrick@wundermanthompson.com',
    license='Apache-2.0',
    python_requires='>=3',
    install_requires=[
        'mkdocs>=1.1.1'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ],
    packages=setuptools.find_packages(),
    entry_points={
        'mkdocs.plugins': [
            "terraform-monorepo = mkdocs_terraform_monorepo_plugin.plugin:TerraformMonorepoPlugin"
        ]
    }
)



import logging
import os
import copy
from pathlib import Path

from mkdocs.utils import warning_filter

log = logging.getLogger(__name__)
log.addFilter(warning_filter)

INCLUDE_STATEMENT = "!tf_modules_root "
# TODO should make these config variables in mkdocs.yml
DOCUMENT_FILE_NAME = 'README.md'
NAVIGATION_KEY_NAME = 'About'


class TfParser:
    def __init__(self, config):
        self.initialNav = config['nav']
        self.config = config
        self.resolvedPaths = []

    def getResolvedPaths(self):
        return self.resolvedPaths

    def resolve(self, nav=None):  # noqa: C901
        if nav is None:
            nav = copy.deepcopy(self.initialNav)
        # log.critical("nav: {} ".format(nav))
        for index, item in enumerate(nav):
            if type(item) is str:
                key = None
                value = item
                # log.critical("value1 = {}".format(value))
            elif type(item) is dict:
                key = list(item.keys())[0]
                value = list(item.values())[0]
                # log.critical("value2 = {}".format(value))
            else:
                key = None
                value = None
            if type(value) is str and INCLUDE_STATEMENT in value:
                # log.critical("value3 = {}".format(value))
                nav[index] = {}
                alias = value.split(INCLUDE_STATEMENT)[0]
                if alias is INCLUDE_STATEMENT:
                    alias == ""
                else:
                    alias = alias.rstrip('/')
                modPath = value.split(INCLUDE_STATEMENT)[-1]
                modNav = CreateModulesNav(
                    self.config,
                    alias,
                    modPath
                ).build()
                # log.critical("modNav = {}".format(modNav.getNav()))
                # log.critical("key = {}".format(key))
                nav[index][key] = modNav.getNav()
                if nav[index][key] is None:
                    del nav[index][key]
                    continue
                else:
                    self.resolvedPaths = [*self.resolvedPaths, *modNav.getResolvedPaths()]
            elif type(value) is list:
                # log.critical("value4 = {}".format(value))
                nav[index] = {}
                nav[index][key] = self.resolve(value)
                if nav[index][key] is None:
                    del nav[index][key]
                    continue
        for index, item in enumerate(nav):
            if nav[index] == {}:
                del nav[index]
        return nav


class CreateModulesNav:
    def __init__(self, config, alias, modulePath):
        # log.critical("config = {}".format(config))
        # log.critical("alias = {}".format(alias))
        # log.critical("modulePath = {}".format(modulePath))
        # log.critical("cwd = {}".format(os.getcwd()))
        # log.critical("config[config_file_path] = {}".format(config['config_file_path']))
        self.rootDir = os.path.normpath(os.path.join(
            os.getcwd(), os.path.dirname(config['config_file_path']), modulePath, '../'))
        # log.critical("rootDir = {}".format(self.rootDir))
        self.alias = alias
        self.modulePath = modulePath
        p = Path(self.modulePath)
        basePath = modulePath
        if len(p.parts) > 1:  # this is if we are in a nested mkdocs for monorepo etc
            basePath = p.relative_to(*p.parts[:1])
        # log.critical("basePath = {}".format(basePath))
        self.absModulePath = os.path.normpath(
            os.path.join(self.rootDir, basePath))
        # log.critical("absModulePath = {}".format(self.absModulePath))
        self.moduleNav = None
        self.resolvedPaths = None

    def getAbsModulePath(self):
        return self.absModulePath

    def build(self):
        if not os.path.exists(self.absModulePath):
            log.critical(
                "[mkdocs-terraform-monorepo] The path {} is not valid. ".format(self.absModulePath) +
                "Please update your 'nav' with a valid path (relative to the mkdocs.yml file).")
            raise SystemExit(1)

        allPaths = []
        for path in Path(self.absModulePath+'/').rglob(DOCUMENT_FILE_NAME):
            allPaths.append(str(path.relative_to(Path(self.absModulePath).parent)))
        # log.critical("allPaths = {}".format(allPaths))
        # TODO need to check full path is not in an ignore list

        allPaths = sorted(allPaths)
        for file in allPaths:
            if self.moduleNav is None:
                self.moduleNav = {}
            self.moduleNav = self.__build_nested(
                os.path.join(self.alias, file), self.moduleNav)
            if self.resolvedPaths is None:
                self.resolvedPaths = []
            self.resolvedPaths.append(
                [os.path.join(self.alias, file), os.path.join(self.rootDir, file)])
        # log.critical("moduleNav = {}".format(self.moduleNav))
        # log.critical("resolvedPaths = {}".format(self.resolvedPaths))
        return self

    def __build_nested_helper(self, path, file, container):
        segs = Path(path).parts
        head, tail = segs[0], segs[1:]
        if len(tail) == 1 or not tail:
            if head == DOCUMENT_FILE_NAME:
                head = NAVIGATION_KEY_NAME
            container[head] = file
            # container = [file]
        else:
            if head not in container:
                container[head] = self.__build_nested_helper(os.path.join(*tail), file, {})
            else:
                if type(container[head]) is str:
                    container[head] = {NAVIGATION_KEY_NAME: container[head]}
                    # container = [container[head]]
                    # send it round again, avoiding the file loop
                container[head] = self.__build_nested_helper(os.path.join(*tail), file, container[head])
        return container

    def __build_nested(self, file, container):
        return self.__build_nested_helper(os.path.join(*(Path(file).parts[1:])), file, container)

    def getNav(self):
        return self.moduleNav

    def getResolvedPaths(self):
        return self.resolvedPaths
